package org.processmining.baseminers.algorithms;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.baseminers.parameters.ConvertTS2PNParameters;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.connections.transitionsystem.TransitionSystemConnection;
import org.processmining.models.graphbased.directed.DirectedGraphElementWeights;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.graphbased.directed.transitionsystem.AcceptStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.StartStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;
import org.processmining.models.graphbased.directed.transitionsystem.TransitionSystem;
import org.processmining.models.semantics.petrinet.Marking;

public class ConvertTS2PNAlgorithm {

	public Petrinet apply(PluginContext context, TransitionSystem transitionSystem,
			ConvertTS2PNParameters parameters) {
		try {
			TransitionSystemConnection connection = context.getConnectionManager()
					.getFirstConnection(TransitionSystemConnection.class, context, transitionSystem);
			return apply(context, transitionSystem, connection.getObjectWithRole(TransitionSystemConnection.WEIGHTS),
					connection.getObjectWithRole(TransitionSystemConnection.STARTIDS),
					connection.getObjectWithRole(TransitionSystemConnection.ACCEPTIDS), parameters);
		} catch (ConnectionCannotBeObtained e) {
			// No start states or accept states available.
			return apply(context, transitionSystem, new DirectedGraphElementWeights(), new StartStateSet(),
					new AcceptStateSet(), parameters);
		}
	}

	public Petrinet apply(PluginContext context, TransitionSystem transitionSystem, DirectedGraphElementWeights weights,
			Collection<Object> startObjects, AcceptStateSet acceptObjects,
			ConvertTS2PNParameters parameters) {
		Petrinet petrinet = PetrinetFactory
				.newPetrinet("Converted from transition system " + transitionSystem.getLabel());
		// Create net.
		Place startPlace = petrinet.addPlace("start");
		Place acceptPlace = petrinet.addPlace("end");
		Map<State, Place> state2PlaceMap = new HashMap<State, Place>();
		org.processmining.models.graphbased.directed.petrinet.elements.Transition petrinetTransition;
		// Convert every state into a place.
		for (State state : transitionSystem.getNodes()) {
			Place place = petrinet.addPlace(state.getIdentifier().toString());
			state2PlaceMap.put(state, place);
		}
		// Convert every transition that exceeds the threshold into a visible transition.
		Set<Place> disconnectedPlaces = new HashSet<Place>(state2PlaceMap.values());
		for (Transition transition : transitionSystem.getEdges()) {
			if (parameters.getThreshold() < weights.get(transition.getSource().getIdentifier(),
					transition.getTarget().getIdentifier(), transition.getIdentifier(), 0)) {
				Place sourcePlace = state2PlaceMap.get(transition.getSource());
				Place targetPlace = state2PlaceMap.get(transition.getTarget());
				petrinetTransition = petrinet.addTransition(transition.getLabel());
				petrinet.addArc(sourcePlace, petrinetTransition);
				petrinet.addArc(petrinetTransition, targetPlace);
				disconnectedPlaces.remove(sourcePlace);
				disconnectedPlaces.remove(targetPlace);
			}
		}
		// Remove disconnected places.
		for (Place place : disconnectedPlaces) {
			petrinet.removePlace(place);
		}
		// Connect surviving start states.
		for (Object startObject : startObjects) {
			Place place = state2PlaceMap.get(transitionSystem.getNode(startObject));
			if (!disconnectedPlaces.contains(place)) {
				petrinetTransition = petrinet.addTransition("");
				petrinetTransition.setInvisible(true);
				petrinet.addArc(startPlace, petrinetTransition);
				petrinet.addArc(petrinetTransition, place);
			}
		}
		// Connect surviving accept states.
		for (Object acceptObject : acceptObjects) {
			Place place = state2PlaceMap.get(transitionSystem.getNode(acceptObject));
			if (!disconnectedPlaces.contains(place)) {
				petrinetTransition = petrinet.addTransition("");
				petrinetTransition.setInvisible(true);
				petrinet.addArc(place, petrinetTransition);
				petrinet.addArc(petrinetTransition, acceptPlace);
			}
		}
		// Add initial and final markings to the net.
		Marking startMarking = new Marking();
		startMarking.add(startPlace);
		context.getProvidedObjectManager().createProvidedObject("Initial marking", startMarking, context);
		context.getConnectionManager().addConnection(new InitialMarkingConnection(petrinet, startMarking));
		Marking acceptMarking = new Marking();
		acceptMarking.add(acceptPlace);
		context.getProvidedObjectManager().createProvidedObject("Final marking", acceptMarking, context);
		context.getConnectionManager().addConnection(new FinalMarkingConnection(petrinet, acceptMarking));
		// Return the net. 
		return petrinet;
	}
}
